import uuid
from datetime import datetime
from typing import List, Optional

from pydantic import Field

from mirix.helpers.datetime_helpers import get_utc_time
from mirix.schemas.mirix_base import MirixBase
from mirix.services.organization_manager import OrganizationManager


class ClientBase(MirixBase):
    __id_prefix__ = "client"


def _generate_client_id() -> str:
    """Generate a random client ID."""
    return f"client-{uuid.uuid4().hex[:8]}"


class Client(ClientBase):
    """
    Representation of a client application.

    Parameters:
        id (str): The unique identifier of the client.
        name (str): The name of the client application.
        status (str): Whether the client is active or not.
        scope (str): Scope of client (read, write, read_write, admin).
        email (str): Optional email for dashboard login.
        password_hash (str): Optional password hash for dashboard login.
        created_at (datetime): The creation date of the client.
    """

    id: str = Field(
        default_factory=_generate_client_id,
        description="The unique identifier of the client.",
    )
    organization_id: Optional[str] = Field(
        OrganizationManager.DEFAULT_ORG_ID,
        description="The organization id of the client",
    )
    name: str = Field(..., description="The name of the client application.")
    status: str = Field("active", description="Whether the client is active or not.")
    scope: str = Field("read_write", description="Scope of client.")

    # Dashboard authentication fields
    email: Optional[str] = Field(None, description="Email address for dashboard login.")
    password_hash: Optional[str] = Field(None, description="Hashed password for dashboard login.")
    last_login: Optional[datetime] = Field(None, description="Last dashboard login time.")

    created_at: Optional[datetime] = Field(default_factory=get_utc_time, description="The creation date of the client.")
    updated_at: Optional[datetime] = Field(default_factory=get_utc_time, description="The update date of the client.")
    is_deleted: bool = Field(False, description="Whether this client is deleted or not.")


class ClientCreate(ClientBase):
    id: Optional[str] = Field(None, description="The unique identifier of the client.")
    name: str = Field(..., description="The name of the client application.")
    status: str = Field("active", description="Whether the client is active or not.")
    scope: str = Field("read_write", description="Scope of client.")
    organization_id: str = Field(..., description="The organization id of the client.")


class ClientUpdate(ClientBase):
    id: str = Field(..., description="The id of the client to update.")
    name: Optional[str] = Field(None, description="The new name of the client.")
    status: Optional[str] = Field(None, description="The new status of the client.")
    scope: Optional[str] = Field(None, description="The new scope of the client.")
    organization_id: Optional[str] = Field(None, description="The new organization id of the client.")



from datetime import datetime
from typing import Any, Dict, List, Optional

from pydantic import Field, field_validator

from mirix.client.utils import get_utc_time
from mirix.constants import MAX_EMBEDDING_DIM
from mirix.schemas.embedding_config import EmbeddingConfig
from mirix.schemas.mirix_base import MirixBase


class KnowledgeVaultItemBase(MirixBase):
    """
    Base schema for knowledge vault items containing common fields.
    """

    __id_prefix__ = "kv_item"
    entry_type: str = Field(..., description="Category (e.g., 'credential', 'bookmark', 'api_key')")
    source: str = Field(..., description="Information on who/where it was provided")
    sensitivity: str = Field(..., description="Data sensitivity level ('low', 'medium', 'high')")
    secret_value: str = Field(..., description="The actual credential or data value")
    caption: str = Field(
        ...,
        description="Description of the knowledge vault item (e.g. 'API key for OpenAI Service')",
    )


class KnowledgeVaultItem(KnowledgeVaultItemBase):
    """
    Representation of a knowledge vault item for storing credentials, bookmarks, etc.

    Additional Parameters:
        id (str): Unique ID for this knowledge vault entry.
        created_at (datetime): Creation timestamp.
        updated_at (Optional[datetime]): Last update timestamp.
    """

    id: Optional[str] = Field(None, description="Unique identifier for the knowledge vault item")
    agent_id: Optional[str] = Field(None, description="The id of the agent this knowledge vault item belongs to")
    client_id: Optional[str] = Field(None, description="The id of the client application that created this item")
    user_id: str = Field(..., description="The id of the user who generated the knowledge vault item")
    created_at: datetime = Field(
        default_factory=get_utc_time,
        description="The creation date of the knowledge vault item",
    )
    updated_at: Optional[datetime] = Field(None, description="The last update date of the knowledge vault item")
    last_modify: Dict[str, Any] = Field(
        default_factory=lambda: {
            "timestamp": get_utc_time().isoformat(),
            "operation": "created",
        },
        description="Last modification info including timestamp and operation type",
    )
    organization_id: str = Field(..., description="The unique identifier of the organization")
    caption_embedding: Optional[List[float]] = Field(None, description="The embedding of the summary")
    embedding_config: Optional[EmbeddingConfig] = Field(
        None, description="The embedding configuration used by the event"
    )

    # NEW: Filter tags for flexible filtering and categorization
    filter_tags: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Custom filter tags for filtering and categorization",
        examples=[
            {"project_id": "proj-abc", "session_id": "sess-xyz", "tags": ["important", "work"], "priority": "high"}
        ],
    )

    # need to validate both details_embedding and summary_embedding to ensure they are the same size
    # NEW: Filter tags for flexible filtering and categorization
    filter_tags: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Custom filter tags for filtering and categorization",
        examples=[
            {"project_id": "proj-abc", "session_id": "sess-xyz", "tags": ["important", "work"], "priority": "high"}
        ],
    )

    @field_validator("caption_embedding")
    @classmethod
    def pad_embeddings(cls, embedding: List[float]) -> List[float]:
        """Pad embeddings to `MAX_EMBEDDING_SIZE`. This is necessary to ensure all stored embeddings are the same size."""
        import numpy as np

        if embedding and len(embedding) != MAX_EMBEDDING_DIM:
            np_embedding = np.array(embedding)
            padded_embedding = np.pad(
                np_embedding,
                (0, MAX_EMBEDDING_DIM - np_embedding.shape[0]),
                mode="constant",
            )
            return padded_embedding.tolist()
        return embedding


class KnowledgeVaultItemCreate(KnowledgeVaultItemBase):
    """
    Schema for creating a new knowledge vault item.

    Inherits all required fields from KnowledgeVaultItemBase.
    """

    pass


class KnowledgeVaultItemUpdate(MirixBase):
    """
    Schema for updating an existing knowledge vault item.

    All fields (except id) are optional so that only provided fields are updated.
    """

    id: str = Field(..., description="Unique ID for this knowledge vault entry")
    agent_id: Optional[str] = Field(None, description="The id of the agent this knowledge vault item belongs to")
    entry_type: Optional[str] = Field(None, description="Category (e.g., 'credential', 'bookmark', 'api_key')")
    source: Optional[str] = Field(None, description="Information on who/where it was provided")
    sensitivity: Optional[str] = Field(None, description="Data sensitivity level ('low', 'medium', 'high')")
    secret_value: Optional[str] = Field(None, description="The actual credential or data value")
    organization_id: Optional[str] = Field(None, description="The unique identifier of the organization")
    updated_at: datetime = Field(default_factory=get_utc_time, description="The update date")
    last_modify: Optional[Dict[str, Any]] = Field(
        None,
        description="Last modification info including timestamp and operation type",
    )
    caption_embedding: Optional[List[float]] = Field(None, description="The embedding of the summary")
    embedding_config: Optional[EmbeddingConfig] = Field(
        None, description="The embedding configuration used by the event"
    )

    filter_tags: Optional[Dict[str, Any]] = Field(
        None, description="Custom filter tags for filtering and categorization"
    )


class KnowledgeVaultItemResponse(KnowledgeVaultItem):
    """
    Response schema for knowledge vault items with additional fields that might be needed by the API.
    """

    pass



from typing import TYPE_CHECKING, Any, Dict, List, Optional

from pydantic import Field, model_validator

from mirix.constants import (
    COMPOSIO_TOOL_TAG_NAME,
    FUNCTION_RETURN_CHAR_LIMIT,
    MIRIX_CORE_TOOL_MODULE_NAME,
    MIRIX_EXTRA_TOOL_MODULE_NAME,
    MIRIX_MEMORY_TOOL_MODULE_NAME,
)
from mirix.schemas.enums import ToolType
from mirix.schemas.mirix_base import MirixBase

try:
    from mirix.functions.functions import (
        derive_openai_json_schema,
        get_json_schema_from_module,
    )
    from mirix.functions.helpers import generate_langchain_tool_wrapper
    from mirix.functions.schema_generator import (
        generate_schema_from_args_schema_v2,
    )

    _FUNCTION_UTILS_AVAILABLE = True
except ImportError:
    derive_openai_json_schema = None  # type: ignore
    get_json_schema_from_module = None  # type: ignore
    generate_langchain_tool_wrapper = None  # type: ignore
    generate_schema_from_args_schema_v2 = None  # type: ignore
    _FUNCTION_UTILS_AVAILABLE = False

if TYPE_CHECKING:
    try:
        from langchain_core.tools import BaseTool as LangChainBaseTool
    except ImportError:
        LangChainBaseTool = Any  # type: ignore


class BaseTool(MirixBase):
    __id_prefix__ = "tool"


class Tool(BaseTool):
    """
    Representation of a tool, which is a function that can be called by the agent.

    Parameters:
        id (str): The unique identifier of the tool.
        name (str): The name of the function.
        tags (List[str]): Metadata tags.
        source_code (str): The source code of the function.
        json_schema (Dict): The JSON schema of the function.

    """

    id: str = BaseTool.generate_id_field()
    tool_type: ToolType = Field(ToolType.CUSTOM, description="The type of the tool.")
    description: Optional[str] = Field(None, description="The description of the tool.")
    source_type: Optional[str] = Field(None, description="The type of the source code.")
    organization_id: Optional[str] = Field(
        None,
        description="The unique identifier of the organization associated with the tool.",
    )
    name: Optional[str] = Field(None, description="The name of the function.")
    tags: List[str] = Field([], description="Metadata tags.")

    # code
    source_code: Optional[str] = Field(None, description="The source code of the function.")
    json_schema: Optional[Dict] = Field(None, description="The JSON schema of the function.")

    # tool configuration
    return_char_limit: int = Field(
        FUNCTION_RETURN_CHAR_LIMIT,
        description="The maximum number of characters in the response.",
    )

    # metadata fields
    created_by_id: Optional[str] = Field(None, description="The id of the user that made this Tool.")
    last_updated_by_id: Optional[str] = Field(None, description="The id of the user that made this Tool.")

    @model_validator(mode="after")
    def populate_missing_fields(self):
        """
        Populate missing fields: name, description, and json_schema.
        """

        if self.tool_type == ToolType.CUSTOM:
            # If it's a custom tool, we need to ensure source_code is present
            if not self.source_code:
                raise ValueError(f"Custom tool with id={self.id} is missing source_code field.")

            # Always derive json_schema for freshest possible json_schema
            # TODO: Instead of checking the tag, we should having `COMPOSIO` as a specific ToolType
            # TODO: We skip this for Composio bc composio json schemas are derived differently
            if COMPOSIO_TOOL_TAG_NAME not in self.tags and _FUNCTION_UTILS_AVAILABLE:
                self.json_schema = derive_openai_json_schema(source_code=self.source_code)
            elif COMPOSIO_TOOL_TAG_NAME not in self.tags and not self.json_schema:
                raise ValueError(
                    "Custom tool is missing json_schema and schema helpers are not available in the client-only build."
                )
        elif self.tool_type in {ToolType.MIRIX_CORE}:
            # If it's mirix core tool, we generate the json_schema on the fly here
            # But only if it's not already provided (e.g., from server)
            if not self.json_schema:
                if not self.name:
                    raise ValueError(f"MIRIX_CORE tool with id={self.id} requires 'name' field to generate schema")
                if get_json_schema_from_module:
                    self.json_schema = get_json_schema_from_module(
                        module_name=MIRIX_CORE_TOOL_MODULE_NAME,
                        function_name=self.name,
                    )
                else:
                    raise ValueError(
                        "Tool schema helpers are not available in the client-only build. Provide json_schema explicitly."
                    )
        elif self.tool_type in {ToolType.MIRIX_MEMORY_CORE}:
            if not self.json_schema:
                if not self.name:
                    raise ValueError(
                        f"MIRIX_MEMORY_CORE tool with id={self.id} requires 'name' field to generate schema"
                    )
                if get_json_schema_from_module:
                    self.json_schema = get_json_schema_from_module(
                        module_name=MIRIX_MEMORY_TOOL_MODULE_NAME,
                        function_name=self.name,
                    )
                else:
                    raise ValueError(
                        "Tool schema helpers are not available in the client-only build. Provide json_schema explicitly."
                    )
        elif self.tool_type in {ToolType.MIRIX_EXTRA}:
            if not self.json_schema:
                if not self.name:
                    raise ValueError(f"MIRIX_EXTRA tool with id={self.id} requires 'name' field to generate schema")
                if get_json_schema_from_module:
                    self.json_schema = get_json_schema_from_module(
                        module_name=MIRIX_EXTRA_TOOL_MODULE_NAME,
                        function_name=self.name,
                    )
                else:
                    raise ValueError(
                        "Tool schema helpers are not available in the client-only build. Provide json_schema explicitly."
                    )
        elif self.tool_type in {ToolType.MIRIX_MCP}:
            # MCP tools have their json_schema already provided by MCP tool registry
            # Skip validation since these are auto-generated tools
            if not self.json_schema:
                raise ValueError(f"MCP tool {self.name} is missing json_schema field")

        # Derive name from the JSON schema if not provided
        if not self.name:
            # TODO: This in theory could error, but name should always be on json_schema
            # TODO: Make JSON schema a typed pydantic object
            self.name = self.json_schema.get("name")

        # Derive description from the JSON schema if not provided
        if not self.description:
            # TODO: This in theory could error, but description should always be on json_schema
            # TODO: Make JSON schema a typed pydantic object
            self.description = self.json_schema.get("description")

        return self


class ToolCreate(MirixBase):
    name: Optional[str] = Field(
        None,
        description="The name of the function (auto-generated from source_code if not provided).",
    )
    description: Optional[str] = Field(None, description="The description of the tool.")
    tags: List[str] = Field([], description="Metadata tags.")
    source_code: str = Field(..., description="The source code of the function.")
    source_type: str = Field("python", description="The source type of the function.")
    json_schema: Optional[Dict] = Field(
        None,
        description="The JSON schema of the function (auto-generated from source_code if not provided)",
    )
    return_char_limit: int = Field(
        FUNCTION_RETURN_CHAR_LIMIT,
        description="The maximum number of characters in the response.",
    )

    @classmethod
    def from_langchain(
        cls,
        langchain_tool: "LangChainBaseTool",
        additional_imports_module_attr_map: dict[str, str] = None,
    ) -> "ToolCreate":
        """
        Class method to create an instance of Tool from a Langchain tool (must be from langchain_community.tools).

        Args:
            langchain_tool (LangChainBaseTool): An instance of a LangChain BaseTool (BaseTool from LangChain)
            additional_imports_module_attr_map (dict[str, str]): A mapping of module names to attribute name. This is used internally to import all the required classes for the langchain tool. For example, you would pass in `{"langchain_community.utilities": "WikipediaAPIWrapper"}` for `from langchain_community.tools import WikipediaQueryRun`. NOTE: You do NOT need to specify the tool import here, that is done automatically for you.

        Returns:
            Tool: A Mirix Tool initialized with attributes derived from the provided LangChain BaseTool object.
        """
        if not _FUNCTION_UTILS_AVAILABLE:
            raise ImportError(
                "LangChain helpers are unavailable in the client-only build. "
                "Install the full Mirix package to enable from_langchain()."
            )

        description = langchain_tool.description
        source_type = "python"
        tags = ["langchain"]
        # NOTE: langchain tools may come from different packages
        wrapper_func_name, wrapper_function_str = generate_langchain_tool_wrapper(
            langchain_tool, additional_imports_module_attr_map
        )
        json_schema = generate_schema_from_args_schema_v2(
            langchain_tool.args_schema, name=wrapper_func_name, description=description
        )

        return cls(
            name=wrapper_func_name,
            description=description,
            source_type=source_type,
            tags=tags,
            source_code=wrapper_function_str,
            json_schema=json_schema,
        )

    @classmethod
    def load_default_langchain_tools(cls) -> List["ToolCreate"]:
        # For now, we only support wikipedia tool
        from langchain_community.tools import WikipediaQueryRun
        from langchain_community.utilities import WikipediaAPIWrapper

        wikipedia_tool = ToolCreate.from_langchain(
            WikipediaQueryRun(api_wrapper=WikipediaAPIWrapper()),
            {"langchain_community.utilities": "WikipediaAPIWrapper"},
        )

        return [wikipedia_tool]

    @classmethod
    def load_default_composio_tools(cls) -> List["ToolCreate"]:
        pass

        # TODO: Disable composio tools for now
        # TODO: Naming is causing issues
        # calculator = ToolCreate.from_composio(action_name=Action.MATHEMATICAL_CALCULATOR.name)
        # serp_news = ToolCreate.from_composio(action_name=Action.SERPAPI_NEWS_SEARCH.name)
        # serp_google_search = ToolCreate.from_composio(action_name=Action.SERPAPI_SEARCH.name)
        # serp_google_maps = ToolCreate.from_composio(action_name=Action.SERPAPI_GOOGLE_MAPS_SEARCH.name)

        return []


class ToolUpdate(MirixBase):
    description: Optional[str] = Field(None, description="The description of the tool.")
    name: Optional[str] = Field(None, description="The name of the function.")
    tags: Optional[List[str]] = Field(None, description="Metadata tags.")
    source_code: Optional[str] = Field(None, description="The source code of the function.")
    source_type: Optional[str] = Field(None, description="The type of the source code.")
    json_schema: Optional[Dict] = Field(
        None,
        description="The JSON schema of the function (auto-generated from source_code if not provided)",
    )
    return_char_limit: Optional[int] = Field(None, description="The maximum number of characters in the response.")

    class Config:
        extra = "ignore"  # Allows extra fields without validation errors
        # TODO: Remove this, and clean usage of ToolUpdate everywhere else


class ToolRunFromSource(MirixBase):
    source_code: str = Field(..., description="The source code of the function.")
    args: Dict[str, Any] = Field(..., description="The arguments to pass to the tool.")
    env_vars: Dict[str, str] = Field(None, description="The environment variables to pass to the tool.")
    name: Optional[str] = Field(None, description="The name of the tool to run.")
    source_type: Optional[str] = Field(None, description="The type of the source code.")
